/*     Bed a menu-driven multi dataformat binary editor for Linux            */
/*     Copyright (C) 1998 Jaap Korthals Altes <binaryeditor+bed@eml.cc>       */
/*                                                                           */
/*     This program is free software; you can redistribute it and/or modify  */
/*     it under the terms of the GNU General Public License as published by  */
/*     the Free Software Foundation; either version 2 of the License, or     */
/*     (at your option) any later version.                                   */
/*                                                                           */
/*     This program is distributed in the hope that it will be useful,       */
/*     but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*     GNU General Public License for more details.                          */
/*                                                                           */
/*     You should have received a copy of the GNU General Public License     */
/*     along with this program; if not, write to the Free Software           */
/*     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.             */
/* Fri Dec  8 22:14:27 2000                                                  */
#ifndef  EDITOR_H
#define EDITOR_H
#include <features.h>
#define HAS_BOOL
#include "screen.h"
#ifdef USETHREADS
#include <pthread.h>
#include "nepsem.h"
#endif
#include "memory.h"
//#include "marks.h"
#include "type.h"
#include "myalloc.h"
/*
#ifdef _WIN32
#include "win.h"
#endif
*/
#define GeenEditor
#define MAXANT 256

const int maxmenus=30;
#define MAXUNDO 5000
struct undo {
	OFFTYPE pos;
	unsigned char *str;
	int len;
	};

class Filter ;
class ScreenPart;

class Empty;
class keylookup;

#ifdef USETHREADS
class Editor;
struct threadgegs {
	ScreenPart *part;
	Editor *ed;
	char *ant;
	int len;
	OFFTYPE pos;
	};
#endif

class Editor {

	public:
		WINDOW *editscreen;
		int fromtop,numberof;

int visualwidth,visualoffset,virtualwidth, visualunit;
// Declaraties
struct Menugeg {
	 const char *name;
	 const char *key;
	int (Editor::*proc)(void);
	};
//Static 
	static  int (Editor::*proc[])(void);
	static  const char *procnames[];

	static  int (Editor::**dynproc)(void);
	static  const char **dynprocnames;
	static int dynprocnr;

	static int procnr;
//DATA:
	Memory mem;
/*s/^\([^(]*\)Editor::/\1 GeenEditor /g

*/
void  GeenEditor enlargeparts(void);
	ScreenPart **parts;
	int maxparts;
	int nrparts;
	int mode;
//	enum {ASCII,HEX} mode;
	char *buffer;
	char *getbuf(void);
	int changed,modified;
	int edge,shownbufsize;
	OFFTYPE filepos ,lastgotopos;
	long editpos,oldpos,updated;
	int dofromfile;
	int commandmode;
	int allheight,Screenheight,Screenwidth,edittop;
	int shouldreinit;
int modechanged(void); 
	int nrx,nry;
	int getcols(void) ;
	int distribute(void) ;

	ScreenPart *searchpart;
#ifdef USETHREADS
	sem_t sema;
	pthread_t threadsearch;
struct  threadgegs gegsthread;
	void stopthread(void); 
	int found;
#endif
#define MAXNUMSTR 12
/* Byte number procedures */
	int maxnum, numbase;
	char numstr[MAXNUMSTR];
	inline int startnum(void) {return 0;};
	int putnumbase(int base);
	void number(void);
	void numline(int y);

	int  GeenEditor nrnumber(void);
	inline int top(void) {return 0;};

int resetnumbase(void) ;
/*
struct block {
	unsigned char *begin;
	int last,free,len,lowest;
	struct block *prev;
	};
*/
//struct block *now,*chunkbefore;


struct  undo undobuf[MAXUNDO];
int lastundo,redos,beginundo,undohalf;
void initundo(void) ;
void delundo(void) ;

unsigned char * GeenEditor expandchunk(unsigned char *ptr,int len,int oldlen) ;

inline struct undo * GeenEditor getnextundo(void) ;


int  GeenEditor popundo(unsigned char **buf,OFFTYPE *pos) ;
int undoone(void) ;
int  GeenEditor strtoundoer(OFFTYPE pos,int size) ;
int  GeenEditor strtoundo(OFFTYPE pos,int size) ;
int  GeenEditor addundo(OFFTYPE pos,unsigned char ch) ;
int  GeenEditor undopos(void)  ;
int  GeenEditor appendstrtoundo(OFFTYPE pos,char *str, int size) ;
int  GeenEditor appendtoundo(OFFTYPE pos, int size) ;
int      GeenEditor extendtoundo(OFFTYPE oldsize);

int	 GeenEditor extendtoundoonly(OFFTYPE oldsize) ;


//PROCEDURES

#undef prodefname
#define prodefname(name,pro)  int pro(void);
#undef prodef
#define prodef(pro)  int pro(void);
#include "procedures.h"
#undef prodef
	int proceskey(void);
int proceskeyonce(void) ;

	Editor(const char *filename);
	~Editor() ;
	int reinit(void);
	int place(int);
//	int changechar(int pos,unsigned char ch);
inline int  GeenEditor printchar(int pos) ;
	int changestr(int pos,unsigned char *str,int size);
	void doexit(void);
	void nocursor(void);
int  GeenEditor fromfile(OFFTYPE pos);
	int fromfile(void) ;
	int geteditpos(void);
	int lowborder(void);

	OFFTYPE getfilepos(void);

/* Geometry */

int  GeenEditor toborderpos(OFFTYPE pos) ;
	int  GeenEditor toborder(OFFTYPE pos) ;
	inline int cols(void) { return nrx;};
	inline int rows(void) { return nry;};

	inline int x(void) { return (editpos%nrx);};
	inline int y(void) { return (editpos/nrx);};
	int writebuf(void);
	void cursorpos(void) ;
	inline void resetpos(void);
	int topos(OFFTYPE pos) ;
int  GeenEditor pretopos(OFFTYPE pos);
	void putline(int y) ;
	int nomessage;
	inline void textovermessage(void);
	inline void restoremessageline(void);
	 void clearmessage(void) ;

	int getinput(const char question[],char *answer) ;
	int inputget(char *answer,const char *format,...) ;

#define NEXTFILE 888
#define PREVFILE 889
#define ASKFILE  890
#define LASTFILE 891
#define FIRSTFILE 891
#define FILENR	892
#define MAXFILES 60


int askmenufile(void) ;

	inline int switchcursor(void) ;
	int nextfast(void) ;

	int save(void);
	int  GeenEditor saveaser(const char *ant) ;

	int extendfile(int len) ;



int  GeenEditor toscreen(void) ;


OFFTYPE  GeenEditor getmem(OFFTYPE pos,char *buf,OFFTYPE len);
OFFTYPE  GeenEditor putmem(OFFTYPE pos,char *buf,OFFTYPE len);



	short searchbytes,searchgrens,indatatype;

FILEHANDLE openfile(const char *ant) ;

FILEHANDLE  GeenEditor tmpopenfile(const char *ant) ;
	char *formatsearch,*searchdataname;

#define SEARCH_REGEX 4
#define SEARCH_FORWARD 1
#define SEARCH_CASE 2
#define SEARCH_BOUNDARY 8
int  GeenEditor searchtype(int newtype) ;
	OFFTYPE dosearch(char *ant,int len);
	OFFTYPE dosearch(OFFTYPE pos, char *ant,int len) ;
	OFFTYPE dosearcher(char *ant,int len);
//	OFFTYPE dosearcher(OFFTYPE pos, char *ant,int len) ;
OFFTYPE  GeenEditor dosearcher(ScreenPart *part,OFFTYPE pos, char *ant,int len) ;
	int searchregex(short yes);
	int searchforward(short yes);
	int searchcase(short yes) ;
	int searchonboundary(short yes) ;
	int searchedge(short yes);
//	int searchlist(char *ant,int len) ;

#if !defined( USETHREADS) || defined(BIOSDISK)
int  GeenEditor searchlist(ScreenPart *part,char *ant, int len) ;
#endif
	int  GeenEditor searcher(int askrepl);
//int  GeenEditor searchlister(char *ant,int len,OFFTYPE pos) ;
int  GeenEditor searchlister(ScreenPart *part,char *ant,int len,OFFTYPE pos) ;
#ifdef USETHREADS

int startthreadsearch(ScreenPart *part,char *ant, int len) ;
#endif
	char *replacestr;
	int replacestrlen;
	int  GeenEditor replacehere(OFFTYPE pos, char *str,int len  ) ;
	int  GeenEditor replacer(char *rep,OFFTYPE foundpos);
//int  GeenEditor replacer(ScreenPart *part,char *rep,OFFTYPE foundpos) ;

	int endedit(void);
	int closefile(void);

	int addbefore(DataType type,int to,int len,int order);
	inline int addafter(DataType type,int to,int len,int order) { return addbefore(type,to+1, len,order); }
 
	int tomode(DataType type,int len,int order) ;
	int  GeenEditor tomodenr(int newmode) ;
//	int tomodeget(void) ;
 void editup(void) ;

	void filesup(void);



//	long digfrom(char *ant); to OFFTYPE??
	inline int maxedit(void) { return ((nrx)*(nry));}

	int screensize(void);
	 void memorize(void);


//	long selects[2];

	static	int	lensel;
	static char *selbuf;
	OFFTYPE other;
	int el;
int  GeenEditor selected(OFFTYPE pos) ;

OFFTYPE  GeenEditor getselregion(OFFTYPE *start) ;
int  GeenEditor setmark(unsigned char *keys,int nr,OFFTYPE pos) ;

keylookup *marksptr;
/*
inline int defaultmark(void) {
	marx.defaultmark(editpos+filepos);
	return 0;
	};
*/
#undef getch
 int getch(void) ;


#define scrollon(x,y) scrollok(x,y)
inline void overline();
/*
inline int output(const char info[]) { }
*/
int output(const char info[]) ;
int message(const char *format ...) ;
 int getkeyinput(const char *format ...) ;

 int  GeenEditor showposition(void) ;



int gotonrwin(int nr) ;


void windel(void) ;



struct menuel {
	int key;
	int (Editor::*proc)(void);
	char *name;
	char *alt;
	char keypos;
	};
struct menu {
	WINDOW *win;
	int len,nu,key;
	struct menuel *el;
	char *name;
	char keypos;
	int maxname;
	int barpos;
	};
void editontop(void) ;

int showmenu(void) ;
static void selectmenu(struct menu *m,int dir) ;
static struct menu *menumk(int x,int el,struct Menugeg pro[], const char *under) ;
static int  GeenEditor menustructs(void) ;
static WINDOW * GeenEditor rewritemenu(struct Editor::menu *m) ;

static int  GeenEditor deletemenu(int menunr) ;
static int  GeenEditor deletemenuitem(int menunr,int pos) ;
static int  GeenEditor addmenuitem(int menunr,const char *name,int pos, int ( Editor::*proc)(void),int insert=0,const char *keystr="") ;
static int  addmenu(const char *name,int menunr,int insert=0) ;


static struct menu *menulist[maxmenus];

static int menuiter,menunu;
//int addtobar(char *name) ;
static int  GeenEditor addtobar(struct Editor::menu *m) ;
static int menus(void) ;
int  GeenEditor fileok(void) ;
int  GeenEditor saveinode(void) ;
int  GeenEditor discard(void) ;
  OFFTYPE  GeenEditor getpart(OFFTYPE begin, OFFTYPE len, char * buf); 
   OFFTYPE  GeenEditor putpart(OFFTYPE begin, OFFTYPE len, char * buf);

OFFTYPE filesize(void);
void erefresh(void);


static int (Editor::*ineditproc[])(void);
static int (Editor::*inmenuproc[])(void);
int  GeenEditor therepeat(int nr) ;
char * GeenEditor getfilename(char *buf);

int  GeenEditor getselbuf(char **buf) ;
int  GeenEditor putselbuf(char *buf,int len) ;

OFFTYPE  GeenEditor getselect(void) ;
OFFTYPE  GeenEditor setselect(OFFTYPE pos) ;

OFFTYPE  GeenEditor posputbuf(OFFTYPE pos,char *buf,OFFTYPE len);
OFFTYPE  GeenEditor putbuf(char *buf,OFFTYPE buflen);


void  GeenEditor splitpos(OFFTYPE pos) ; 

int  GeenEditor menuresized(void);
int  GeenEditor getmode(DataType type,int len,int order) ;
int  GeenEditor installpart(ScreenPart *sub) ;
int  GeenEditor installpart(ScreenPart *sub,int separ) ;

int sizelabels(ScreenPart *sub,int kind,int nr,...) ;
int sizelabels(ScreenPart *sub,int kind,int nr,va_list ap,bool) ;
int  GeenEditor forcetruncate(void) ;
int  GeenEditor filetruncate(OFFTYPE vanaf);
int  GeenEditor savesel(const char *ant) ;
//int  GeenEditor saveshownname(const char *ant) ;

ssize_t saveshownname(const char *ant) ;
int  GeenEditor readshownname(const char *ant) ;
//OFFTYPE formsearch(OFFTYPE from,char *ant,short len) ;
OFFTYPE  GeenEditor formsearch(ScreenPart *datashown,OFFTYPE from,char *ant,short len) ;
OFFTYPE  GeenEditor memnextsearch(void) ;
//OFFTYPE  GeenEditor memnextsearch(ScreenPart *datashown) ;
//OFFTYPE  GeenEditor memsearch(OFFTYPE pos,char *str,int len) ;
OFFTYPE  GeenEditor memsearch(ScreenPart *part,OFFTYPE pos,char *str,int len) ;

//int  GeenEditor nextsearching(ScreenPart *part) ;
ScreenPart *activemode(void) ;

int  GeenEditor getanswer(const char *question,char *answ,int anslen,int ansroom=5) ;
int  GeenEditor addtocontain(ScreenPart *part) ;
BLOCKTYPE blocksize(void);

int  GeenEditor addfilter(Filter *fil) ;
int  GeenEditor playmacrofile(const char *name) ;
int  GeenEditor playmacro(char *buf,int len);

int  GeenEditor savecontain(void) ;
FILEHANDLE  GeenEditor getsavedatahandler(const char *question) ;
int  GeenEditor savealldatatypes(void) ;
int  GeenEditor loaddatatype(char *gegs,int endf,int addfield) ;
int  GeenEditor proconname(char *str) ;
int  GeenEditor configASCII(ScreenPart *sub,int kind);
int  GeenEditor partconfig(ScreenPart *sub,int kind=1) ;
ScreenPart * GeenEditor replacepart(ScreenPart *part,int pos=-1) ;
int  GeenEditor makevisible(ScreenPart *part,int pos) ;
OFFTYPE  GeenEditor nextregexshown(ScreenPart *datashown);
int  GeenEditor emptyconfig(Empty *emp,int kind) ;
private:
/*
inline void showerror(void) {
#ifdef _WIN32
	const int meslen=200;
	char mes[meslen];
	if(winerror(mes,meslen))
		output(mes);
#else
	output(strerror(errno));
#endif
	}
*/
//					    pos=(mem.lastforward&1)?(edge?0:filepos+editpos+1): (edge?(mem.size()-1):(filepos+editpos-1));
inline OFFTYPE searchstartpos() {return (mem.lastforward&1)?(edge?0:filepos+editpos+1): (edge?(mem.size()-1):(filepos+editpos-1)); }
int ismessage;
int  GeenEditor showmessage(void) ;
int  GeenEditor realoutput(const char info[]) ;
OFFTYPE *previouspos;
int previouspositer, previousposhead;
public:
OFFTYPE getpreviouspos(void);
char justopened;
int  GeenEditor insertshown(char *gegs, off_t endf) ;
#define mkmenunum(num) int GeenEditor menu##num##show(void);
#include "mkmenus.h"
#undef mkmenunum
};
#endif
